import numpy as np
from scipy.spatial.distance import pdist, squareform
from mppy.stress import calculate_kruskal_stress
import time


def force_2d(matrix, max_iter=50, delta_frac=8.0, eps=1e-6):
    """

    :param matrix:
    :param max_iter:
    :param delta_frac:
    :param eps:
    :return:
    """
    orig_matrix = matrix
    # data_matrix = matrix[:, :-1]
    data_matrix = orig_matrix.copy()
    instances = orig_matrix.shape[0]
    dimensions = orig_matrix.shape[1] - 1
    # clusters = orig_matrix[:, dimensions]
    # matrix_2d = np.random.random((instances, 2))
    # clusters = clusters.astype(int)

    start_time = time.time()
    matrix_2d = _force(data_matrix, max_iter, delta_frac, eps)
    print("Algorithm execution: %s seconds" % (time.time() - start_time))
    print("Stress: %s" % calculate_kruskal_stress(data_matrix, matrix_2d))

    return matrix_2d


def _force(matrix, max_iter=50, delta_frac=80, eps= 1e-6):
    """

    :param matrix:
    :param max_iter:
    :param delta_frac:
    :param eps:
    :return:
    """
    matrix_2d = np.random.random((matrix.shape[0], 2))
    distance_matrix = squareform(pdist(matrix))
    index = np.random.permutation(matrix.shape[0])
    for i in range(max_iter):
        for i in range(matrix.shape[0]):
            instance1 = index[i]
            for j in range(matrix.shape[0]):
                instance2 = index[j]

                if instance1 == instance2:
                    continue
                else:
                    x1x2 = matrix_2d[instance2, 0] - matrix_2d[instance1, 0]
                    y1y2 = matrix_2d[instance2, 1] - matrix_2d[instance1, 1]
                    dr2 = np.hypot(x1x2, y1y2)
                    # dr2 = np.sqrt((x1x2 * x1x2) + (y1y2 * y1y2))

                if dr2 < eps:
                    dr2 = eps

                drn = distance_matrix[instance1, instance2] - dr2
                delta = drn - dr2
                delta /= delta_frac

                matrix_2d[instance2, 0] += delta * (x1x2 / dr2)
                matrix_2d[instance2, 1] += delta * (y1y2 / dr2)

    return matrix_2d



import numpy as np
import scipy as sp
from mppy.stress import calculate_kruskal_stress
from mppy.force import _force
import time


def lamp_2d(matrix, sample_indices=None, initial_sample=None, proportion=1):
    orig_matrix = matrix
    # data_matrix = matrix[:, :-1]
    data_matrix = orig_matrix.copy()
    instances = orig_matrix.shape[0]
    dimensions = orig_matrix.shape[1] - 1
    matrix_2d = np.random.random((instances, 2))

    start_time = time.time()
    if sample_indices is None:
        sample_indices = np.random.randint(0, instances - 1, int(1.0 * np.sqrt(instances)))
        initial_sample = None

    if initial_sample is None:
        aux = data_matrix[sample_indices, :]
        initial_sample = _force(aux)

    Xs = np.array((data_matrix[sample_indices, :]))
    for i in range(instances):
        point = np.array(data_matrix[i, :])

        #calculating alphas
        skip = False
        alphas = np.zeros((len(initial_sample)))
        for j in range(len(initial_sample)):
            dist = np.sum(sp.square(Xs[j,:] - point))
            if dist < 1e-6:
                #ponto muito perto do sample point
                matrix_2d[i, :] = initial_sample[j, :]
                skip = True
                break

            alphas[j] = 1.0 / dist

        if skip is True:
            continue

        c = len(initial_sample) * proportion
        if c < len(initial_sample):
            index = alphas[np.argsort(-temp)]
            j = c
            for j in range(len(initial_sample)):
                alphas[index[j]] = 0
        alphas_sum = np.sum(alphas)

        #calculate \til{x} and \til{Y}
        #print(alphas.shape, Xs.shape, matrix.initial_sample.shape, alphas_sum.shape)
        Xtil = np.dot(alphas, Xs) / alphas_sum
        Ytil = np.dot(alphas, initial_sample) / alphas_sum

        #calculate \hat{X} and \hat{Y}
        Xhat = Xs
        Xhat[:, ] -= Xtil
        Yhat = initial_sample
        Yhat[:,] -= Ytil

        d = np.dot(np.transpose(Xhat),Yhat)
        U, s, V = sp.linalg.svd(np.dot(np.transpose(Xhat),Yhat))
        #aux = np.zeros((dimensions, matrix_2d.shape[1]))
        aux = np.zeros((orig_matrix.shape[1], matrix_2d.shape[1]))

        for k in range(matrix_2d.shape[1]):
            aux[k, range(matrix_2d.shape[1])] = V[k]

        M = np.dot(U, aux)
        matrix_2d[i] = np.dot((point - Xtil), M) + Ytil

    print("Algorithm execution: %s seconds" % (time.time() - start_time))
    print("Stress: %s" % calculate_kruskal_stress(data_matrix, matrix_2d))

    return matrix_2d


import numpy as np
from scipy.spatial.distance import pdist, squareform


def calculate_kruskal_stress(distance_rn, distance_r2):
    """
    distance_r2 = squareform(pdist(distance_r2))
    distance_rn = squareform(pdist(distance_rn))
    num = np.sum(np.power(distance_r2 - distance_rn, 2))
    den = np.sum(np.power(distance_rn, 2))
    """

    distance_rn = squareform(pdist(distance_rn))
    distance_r2 = squareform(pdist(distance_r2))

    num = 0.0
    den = 0.0
    for i in range(distance_rn.shape[0]):
        for j in range(1, distance_r2.shape[0]):
            if i < j:
                dist_rn = distance_rn[i,j]
                dist_r2 = distance_r2[i,j]

                num += (dist_rn - dist_r2) * (dist_rn - dist_r2)
                den += dist_rn * dist_rn

    return num / den


import numpy as np
import scipy as sp
from scipy.spatial.distance import pdist, squareform
import mppy.force as force
import mppy.sammon as sammon
from mppy.stress import calculate_kruskal_stress
import time


def pekalska_2d(matrix, sample_indices=None, inital_sample=None):
    orig_matrix = matrix
    # data_matrix = matrix[:, :-1]
    data_matrix = orig_matrix.copy()
    instances = orig_matrix.shape[0]
    dimensions = orig_matrix.shape[1] - 1
    matrix_2d = np.random.random((instances, 2))

    start_time = time.time()
    if sample_indices is None:
        sample_indices = np.random.randint(0, instances - 1, int(3.0 * np.sqrt(instances)))

    Ds = data_matrix[sample_indices, :]
    if inital_sample is None:
        #inital_sample = force._force(Ds)
        inital_sample = sammon._sammon(Ds)

    n_rows, n_cols = inital_sample.shape

    # criando base D
    D = np.zeros((n_rows, n_rows))
    D = squareform(pdist(Ds))

    #criando base Y
    Y = np.zeros((n_rows, 2))
    Y = inital_sample.copy()

    #encontrar o V
    P, L, U = sp.linalg.lu(D)
    #L = sp.linalg.cholesky(D)
    result = np.linalg.solve(L, Y)
    V = np.transpose(result)

    #calculando a projeção
    for i in range(instances):
        row = data_matrix[i, :]
        dists = np.zeros((inital_sample.shape[0]))

        for j in range(len(dists)):
            dists[j] = np.linalg.norm(row - Ds[j,:])

        matrix_2d[i,0] = np.dot(dists, V[0,:])
        matrix_2d[i,1] = np.dot(dists, V[1,:])

    print("Algorithm execution: %s seconds" % (time.time() - start_time))
    print("Stress: %s" % calculate_kruskal_stress(data_matrix, matrix_2d))

    return matrix_2d



import numpy as np
from mppy.force import _force
from scipy.spatial.distance import squareform, pdist
from mppy.stress import calculate_kruskal_stress
import time


def lsp_2d(matrix, sample_indices=None, sample_data=None, k=15, dim=2):
    """
    Least Square Projection
    :param matrix: a high-dimensional matrix dataset
    :param sample_indices: Samples indices used as control points
    :param sample_data: data samples of original dataset
    :param neighbors: number of neighbors
    :param dim: final dimension of the projection
    :return:
    """
    orig_matrix = matrix
    instances = orig_matrix.shape[0]
    dimensions = orig_matrix.shape[1] - 1
    matrix_2d = np.random.random((instances, 2))
    #data_matrix = matrix[:, :-1]
    data_matrix = orig_matrix.copy()

    start_time = time.time()
    if sample_indices is None:
        sample_indices = np.random.randint(0, instances-1, int(1.0 * np.sqrt(instances)))
        sample_data = None

    if sample_data is None:
        aux = data_matrix[sample_indices, :]
        sample_data = _force(aux)

    nc = sample_data.shape[0]
    A = np.zeros((instances+nc, instances))
    Dx = squareform(pdist(data_matrix))
    for i in range(instances):
        neighbors = np.argsort(Dx[i, :])[1:k + 1]
        A[i, i] = 1.0
        alphas = Dx[i, neighbors]
        if any(alphas < 1e-9):
            alphas[np.array([idx for idx, item in enumerate(alphas) if item < 1e-9])] = 1
            alphas = 0
        else:
            alphas = 1 / alphas
            alphas = alphas / np.sum(alphas)
            alphas = alphas / np.sum(alphas)

        A[i, neighbors] = -alphas


    nc = sample_indices.shape[0]
    A = np.zeros((instances+nc, instances))

    Dx = squareform(pdist(data_matrix))
    for i in range(instances):
        neighbors = np.argsort(Dx[i, :])[1:k+1]
        A[i,i] = 1.0
        alphas = Dx[i, neighbors]
        if any(alphas < 1e-9):
            alphas[np.array([idx for idx, item in enumerate(alphas) if item < 1e-9])] = 1
            alphas = 0
        else:
            alphas = 1 / alphas
            alphas = alphas / np.sum(alphas)
            alphas = alphas / np.sum(alphas)
        A[i, neighbors] = -alphas

    count = 0
    for i in range(instances, A.shape[0]):
        A[i, sample_indices[count]] = 1.0
        count = count + 1

    b = np.zeros((instances+nc, 2))

    for j in range(sample_data.shape[0]):
        b[j+instances, 0] = sample_data[j,0]
        b[j+instances, 1] = sample_data[j,1]

    X = np.linalg.inv(np.dot(A.transpose(),A))
    Y = np.dot(np.transpose(A), b)
    matrix_2d = np.dot(X,Y)

    print("Algorithm execution: %s seconds" % (time.time() - start_time))
    print("Stress: %s" % calculate_kruskal_stress(data_matrix, matrix_2d))

    return matrix_2d


import matplotlib.pyplot as mplpy
import numpy as np

def simple_scatter_plot(matrix_2d, clusters=None):
    if clusters is None:
        clusters = np.zeros((matrix_2d.shape[0]))
    mplpy.scatter(matrix_2d[:,0], matrix_2d[:,1], c=clusters)
    mplpy.show()

def semi_interactive_scatter_plot(matrix_2d, matrix, clusters=None):
    if clusters is None:
        clusters = np.zeros((matrix_2d.shape[0]))

    if 1:  # picking on a scatter plot (matplotlib.collections.RegularPolyCollection)

        def onpick3(event):
            ind = event.ind
            print("Element: ",ind,"->", matrix[ind,:])

        fig = mplpy.figure()
        ax1 = fig.add_subplot(111)
        ax1.scatter(matrix_2d[:,0], matrix_2d[:,1], c=clusters, picker=True)
        fig.canvas.mpl_connect('pick_event', onpick3)

    mplpy.show()


import numpy as np
import scipy as sp
from sklearn.preprocessing import scale
import mppy.force as force
import mppy.sammon as sammon
from mppy.stress import calculate_kruskal_stress
import time


def plmp_2d(matrix, sample_indices=None, sample_data=None, dim=2):
    orig_matrix = matrix
    # data_matrix = matrix[:, :-1]
    data_matrix = orig_matrix.copy()
    instances = orig_matrix.shape[0]
    dimensions = orig_matrix.shape[1]
    initial_matrix = np.random.random((dimensions, dim))

    start_time = time.time()
    if sample_indices is None:
        sample_indices = np.random.randint(0, instances - 1, int(1.0 * np.sqrt(instances)))

    Xs = data_matrix[sample_indices, :]
    if sample_data is None:
        aux = data_matrix[sample_indices, :]
        #sample_data = force._force(aux)
        sample_data = sammon._sammon(aux)

    L = np.transpose(Xs)
    for i in range(dim):
        A = np.dot(L, np.transpose(L))
        B = np.linalg.inv(A)
        C = np.dot(np.transpose(L), B)
        D = np.dot(np.transpose(sample_data[:, i]), C)
        initial_matrix[:, i] = D

    project = np.zeros((instances, dim))
    for j in range(instances):
        project[j, :] = np.dot(data_matrix[j, :], initial_matrix)

    print("Algorithm execution: %s seconds" % (time.time() - start_time))
    print("Stress: %s" % calculate_kruskal_stress(data_matrix, project))

    return project



from distutils.core import setup
setup(
    name = 'mppy',
    packages = ['mppy'], # this must be the same as the name above
    version = '0.3',
    description = 'Multidimensional Projection in Python',
    author = 'Thiago Henrique Ferreira',
    author_email = 'thiagohferreira10@gmail.com',
    url = 'https://github.com/thiagohenriquef/mppy',
    download_url = 'https://github.com/thiagohenriquef/mppy/archive/0.3.tar.gz', # I'll explain this in a second
    keywords = ['multidimensional projection', 'data visualization', 'dimensionality reduction'], # arbitrary keywords
    classifiers=['Development Status :: 4 - Beta','Programming Language :: Python :: 3'],
    install_requires = ['numpy', 'scipy', 'matplotlib', 'sklearn'],
)


import numpy as np
from scipy.spatial.distance import pdist, squareform, euclidean
from mppy.stress import calculate_kruskal_stress
from sklearn import manifold
import time


def sammon(matrix, initial_projection=None, max_iter=50, magic_factor=0.3, tol=1e-4, dim=2):
    orig_matrix = matrix
    # data_matrix = matrix[:, :-1]
    data_matrix = orig_matrix.copy()
    distance_matrix = squareform(pdist(data_matrix))

    start_time = time.time()
    matrix_2d = _sammon(matrix, initial_projection, max_iter, magic_factor, tol, dim)
    print("Algorithm execution: %s seconds" % (time.time() - start_time))
    print("Stress: %s" % calculate_kruskal_stress(data_matrix, matrix_2d))

    return matrix_2d


def _sammon(data_matrix, initial_projection=None, max_iter=50, magic_factor=0.3, tol=1e-4, dim=2):
    # data_matrix = data_matrix[:, :-1]
    distance_matrix = squareform(pdist(data_matrix))

    if initial_projection is None:
        mds = manifold.MDS(n_components=dim, dissimilarity="euclidean")
        initial_projection = mds.fit_transform(data_matrix)

    for i in range(max_iter):
        sum_dist_rn = 0.0

        projection_aux = initial_projection.copy()
        for a in range(distance_matrix.shape[0]):
            for b in range(distance_matrix.shape[0]):
                if distance_matrix[a,b] < tol:
                    distance_matrix[a,b] = tol

                sum_dist_rn += distance_matrix[a,b]

        c = -2 / sum_dist_rn
        #print(c)

        for p in range(distance_matrix.shape[0]):
            for q in range(dim):
                sum_inder_1 = 0.0
                sum_inder_2 = 0.0

                for j in range(distance_matrix.shape[0]):
                    if j != p:
                        x1x2 = projection_aux[p,0] - projection_aux[j,0]
                        y1y2 = projection_aux[p,1] - projection_aux[j,1]
                        dist_pj = np.sqrt(abs(x1x2 * x1x2 + y1y2 * y1y2))

                        if dist_pj < tol:
                            dist_pj = tol

                        sum_inder_1 += ((distance_matrix[p,j] - dist_pj) /
                                        (distance_matrix[p,j] * dist_pj)) * (initial_projection[p, q] - initial_projection[j, q])

                        sum_inder_2 += (1 / (distance_matrix[p,j] * dist_pj)) * \
                                       ((distance_matrix[p,j] - dist_pj) -
                                        ((np.power((initial_projection[p, q] - initial_projection[j, q]), 2) / dist_pj) *
                                         (1 + ((distance_matrix[p,j] - dist_pj) / dist_pj))))

                delta_pq = ((c * sum_inder_1) / abs(c * sum_inder_2))
                initial_projection[p, q] -= magic_factor * delta_pq
    return initial_projection